/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                           Psp.c                         **/
/**                                                         **/
/** This file contains PSP-dependent subroutines and        **/
/** drivers. It includes common drivers from Common.h.      **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994-2005                 **/
/**               Elan Feingold   1995                      **/
/**               Ville Hallik    1996                      **/
/**               Akop Karapetyan 2007                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#ifdef PSP

/** Private #includes ****************************************/
#include "msx/MSX.h"
#include "msx/Sound.h"

#include "psp/video.h"
#include "psp/ctrl.h"
#include "psp/kybd.h"
#include "MenuPsp.h"

/** PSP SDK includes *****************************************/
#include <pspkernel.h>
#include <pspdisplay.h>
#include <pspdebug.h>
#include <psppower.h>
#include <pspgu.h>
#include <psprtc.h>

/** Standard Unix/X #includes ********************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

/** Public parameters ****************************************/
int UseSound  = 44100;          /* Sound driver frequency    */
extern int UpdateFreq;
extern int VSync;

/** Various variables ****************************************/
#define WIDTH  272
#define HEIGHT 228

#define JST_SET(K)  JoyState|=K
#define JST_RES(K)  JoyState&=~K

typedef unsigned short pixel;

static unsigned int BPal[256],XPal[80],XPal0; 
static pixel *XBuf;
static byte JoyState;
static int CurDisk[2];

/** Sound-related definitions ********************************/
#ifdef SOUND
static int SndSwitch = (1<<MAXCHANNELS)-1;
static int SndVolume = 255/MAXCHANNELS;
#endif

extern struct GameConfig GameConfig;
extern int DisplayMode;
extern int Frameskip;
extern PspImage *Screen;

static int ScreenX;
static int ScreenY;
static int ScreenW;
static int ScreenH;
static int ShowKeyboard;

static int FlashMenu;
static int TicksPerUpdate;
static u64 LastTick;
static u64 CurrentTick;
int Frame;

#define BUTTON_MAPPINGS 14

static const u64 ButtonMask[BUTTON_MAPPINGS] = 
{
  PSP_CTRL_UP,       PSP_CTRL_DOWN,
  PSP_CTRL_LEFT,     PSP_CTRL_RIGHT,
  PSP_CTRL_SQUARE,   PSP_CTRL_CROSS,
  PSP_CTRL_CIRCLE,   PSP_CTRL_TRIANGLE,
  PSP_CTRL_LTRIGGER, PSP_CTRL_RTRIGGER,
  PSP_CTRL_SELECT,   PSP_CTRL_START,
  PSP_CTRL_LTRIGGER | PSP_CTRL_RTRIGGER, 
  PSP_CTRL_START | PSP_CTRL_SELECT
};
static const int ButtonMapId[BUTTON_MAPPINGS] = 
{
  MAP_BUTTON_UP,       MAP_BUTTON_DOWN,
  MAP_BUTTON_LEFT,     MAP_BUTTON_RIGHT,
  MAP_BUTTON_SQUARE,   MAP_BUTTON_CROSS,
  MAP_BUTTON_CIRCLE,   MAP_BUTTON_TRIANGLE,
  MAP_BUTTON_LTRIGGER, MAP_BUTTON_RTRIGGER,
  MAP_BUTTON_SELECT,   MAP_BUTTON_START,
  MAP_BUTTON_LRTRIGGERS,
  MAP_BUTTON_STARTSELECT
};

static void OpenMenu();
static void ResetInput();
static void ToggleInput(unsigned int code, int on);
static void HandleSpecialInput(int code, int on);

/** InitMachine() ********************************************/
/** Allocate resources needed by Unix/X-dependent code.     **/
/*************************************************************/
int InitMachine(void)
{
  int J,I;

  /* Initialize keyboard */
  pspKybdCreate();

  /* Initialize menu */
  InitMenu();
  FlashMenu=1;

  /* Reset all variables */
  JoyState=0x00;
  CurDisk[0]=CurDisk[1]=0;
  ShowKeyboard=0;

  /* PSP settings */
  pspVideoSetFilter(PSP_VIDEO_LINEAR);

  /* Screen buffer */
  XBuf=Screen->Pixels;

  /* Reset the palette */
  for(J=0;J<16;J++) XPal[J]=0;
  XPal0=0;

  /* Set SCREEN8 colors */
  for(J=0;J<64;J++)
  {
    I=(J&0x03)+(J&0x0C)*16+(J&0x30)/2;
    XPal[J+16]=RGB16((J&0x30)*255/48,(J&0x0C)*255/12,(J&0x03)*255/3);
    BPal[I]=BPal[I|0x04]=BPal[I|0x20]=BPal[I|0x24]=XPal[J+16];
  }

#ifdef SOUND
  /* Initialize sound */   
  if(InitSound(UseSound,Verbose))
    SetChannels(SndVolume,SndSwitch);
#endif

  /* Init screen position */
  ScreenW = Screen->Width;
  ScreenH = Screen->Height;
  ScreenX = ScreenY = 0;

  Frame = 0;

  return(1);
}

/** TrashMachine() *******************************************/
/** Deallocate all resources taken by InitMachine().        **/
/*************************************************************/
void TrashMachine(void)
{
#ifdef SOUND
  StopSound();
  TrashSound();
#endif

  /* Destroy keyboard */
  pspKybdDestroy();

  /* Destroy menu */
  TrashMenu();
}

/** PutImage() ***********************************************/
/** Put an image on the screen.                             **/
/*************************************************************/
void PutImage(void)
{
  /* Skip frames, if necessary */
  if (++Frame <= Frameskip) return;
  Frame = 0;

  /* Draw the screen */
  pspVideoPutImage(Screen, ScreenX, ScreenY, ScreenW, ScreenH);

  /* Show keyboard */
  if (ShowKeyboard)
  {
    pspKybdShow(); //SCR_WIDTH / 2, SCR_HEIGHT / 2);
  }

  /* Wait if needed */
  if (UpdateFreq)
  {
    while (CurrentTick - LastTick < TicksPerUpdate)
      sceRtcGetCurrentTick(&CurrentTick);
    LastTick = CurrentTick;
  }

  /* Wait for VSync signal */
  if (VSync) pspVideoWaitVSync();

  /* Swap buffers */
  pspVideoSwapBuffers();
}


/** GetKeyboardStatus() **************************************/
/** Gets status of specific MSX key (set/unset).            **/
/*************************************************************/
static int GetKeyboardStatus(unsigned int code)
{
  return !(KeyMap[Keys[code][0]] & Keys[code][1]);
}

/** HandleKeyboardInput() ************************************/
/** Modify MSX keyboard matrix.                             **/
/*************************************************************/
static inline void HandleKeyboardInput(unsigned char code, int on)
{
  if (on) KBD_SET(code);
  else KBD_RES(code);
}

/** Keyboard() ***********************************************/
/** Check for keyboard events, parse them, and modify MSX   **/
/** keyboard/joystick, or handle special keys               **/
/*************************************************************/
void Keyboard(void)
{
  int i;

  /* If starting up, flash the menu */
  if (FlashMenu)
  {
    FlashMenu=0;

    /* Unload all carts (fMSX bug?) */
    for (i = 0; i < MAXCARTS; i++)
      LoadCart(NULL, i);

    /* Display the menu */
    OpenMenu();
  }

  /* Clear joystick state */
  JoyState=0x00;

  /* Check the inputs */
  SceCtrlData pad;
  if (pspCtrlPollControls(&pad))
  {
    if (ShowKeyboard)
      pspKybdNavigate(&pad, HandleKeyboardInput);

    ToggleInput(GameConfig.ButtonMap[MAP_ANALOG_UP], pad.Ly<32);
    ToggleInput(GameConfig.ButtonMap[MAP_ANALOG_DOWN], pad.Ly>=224);
    ToggleInput(GameConfig.ButtonMap[MAP_ANALOG_LEFT], pad.Lx<32);
    ToggleInput(GameConfig.ButtonMap[MAP_ANALOG_RIGHT], pad.Lx>=224);

    for (i = 0; i < BUTTON_MAPPINGS; i++)
      ToggleInput(GameConfig.ButtonMap[ButtonMapId[i]], (pad.Buttons & ButtonMask[i]) == ButtonMask[i]);
  }
}

/** Joystick() ***********************************************/
/** Query position of a joystick connected to port N.       **/
/** Returns 0.0.F2.F1.R.L.D.U.                              **/
/*************************************************************/
unsigned int Joystick(void) { return(JoyState); }

/** Mouse() **************************************************/
/** Query coordinates of a mouse connected to port N.       **/
/** Returns F2.F1.Y.Y.Y.Y.Y.Y.Y.Y.X.X.X.X.X.X.X.X.          **/
/*************************************************************/
unsigned int Mouse(byte N) { return 0; }

/** SetColor() ***********************************************/
/** Set color N (0..15) to R,G,B.                           **/
/*************************************************************/
void SetColor(register byte N,register byte R,register byte G,register byte B)
{
  if (N) XPal[N]=RGB16(R, G, B);
  else XPal0=RGB16(R, G, B);
}

/** ResetInput() *********************************************/
/** Clear joystick/keyboard matrix                          **/
/*************************************************************/
static void ResetInput()
{
  JoyState=0x00;
  memset((void*)KeyMap, 0xFF, sizeof(KeyMap));
}

/** OpenMenu() ***********************************************/
/** Displays fMSX-PSP menu                                  **/
/*************************************************************/
static void OpenMenu()
{
  /* Stop sound & display menu */
  StopSound();
  DisplayMenu();

  float ratio;

  /* Recompute screen size/position */
  switch (DisplayMode)
  {
  default:
  case DISPLAY_MODE_UNSCALED:
    ScreenW = Screen->Width;
    ScreenH = Screen->Height;
    break;
  case DISPLAY_MODE_FIT_HEIGHT:
    ratio = (float)SCR_HEIGHT / (float)Screen->Height;
    ScreenW = (float)Screen->Width * ratio;
    ScreenH = SCR_HEIGHT;
    break;
  case DISPLAY_MODE_FILL_SCREEN:
    ScreenW = SCR_WIDTH;
    ScreenH = SCR_HEIGHT;
    break;
  }

  ScreenX = (SCR_WIDTH / 2) - (ScreenW / 2);
  ScreenY = (SCR_HEIGHT / 2) - (ScreenH / 2);

  /* Recompute update frequency */
  if (UpdateFreq)
  {
    int TickResolution = sceRtcGetTickResolution();
    TicksPerUpdate = TickResolution / (UpdateFreq / (Frameskip + 1));
    sceRtcGetCurrentTick(&LastTick);
  }

  Frame = 0;
  ShowKeyboard = 0;

  /* Reinitialize input matrix & resume sound */
  ResetInput();
  ResumeSound();
}

/** ToggleInput() ********************************************/
/** Modify keyboard/joystick matrix based on provided code  **/
/** toggle status                                           **/
/*************************************************************/
static void ToggleInput(unsigned int code, int on)
{
  if (code & KBD && !ShowKeyboard)
  {
    HandleKeyboardInput(CODE_MASK(code), on);
  }
  else if (code & JST && !ShowKeyboard)
  {
    if (on) JST_SET(CODE_MASK(code));
    else JST_RES(CODE_MASK(code));
  }
  else if (code & SPC)
  {
    HandleSpecialInput(CODE_MASK(code), on);
  }
}

/** HandleSpecialInput() *************************************/
/** Handles emulator-specific buttons                       **/
/*************************************************************/
static void HandleSpecialInput(int code, int on)
{
  switch (code)
  {
  case SPC_MENU:
    if (on) OpenMenu();
    break;
  case SPC_KYBD:
    if (ShowKeyboard != on)
    {
      if (on)
      {
        pspKybdReinit(GetKeyboardStatus);
      }
      else
      {
        /* Clear the buffers */
        pspVideoClearScreen();
        pspVideoSwapBuffers();
        pspVideoClearScreen();
      }
    }
    ShowKeyboard = on;
    break;
  }
}

/** Part of the code common for Unix/X and MSDOS drivers *****/ 
#include "Common.h"

#endif /* PSP */
