/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                          SndPsp.c                       **/
/**                                                         **/
/** This file contains standard sound generation routines   **/
/** for Playstation Portable console                        **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996-2007                 **/
/**               Akop Karapetyan 2007                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#include "msx/Sound.h"

typedef struct {
        short l, r;
} sample_t;

int MasterVolume = 200;                  /* These are made public so  */
int MasterSwitch = (1<<SND_CHANNELS)-1;  /* that LibWin.c can access  */
static int SndRate      = 0;
static int NoiseGen;
static int Wave[SND_BUFSIZE];

static struct
{
  int Type;                       /* Channel type (SND_*)             */
  int Freq;                       /* Channel frequency (Hz)           */
  int Volume;                     /* Channel volume (0..255)          */

  const signed char *Data;        /* Wave data (-128..127 each)       */
  int Length;                     /* Wave length in Data              */
  int Rate;                       /* Wave playback rate (or 0Hz)      */
  int Pos;                        /* Wave current position in Data    */

  int Count;                      /* Phase counter                    */
} CH[SND_CHANNELS] =
{
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 },
  { SND_MELODIC,0,0,0,0,0,0 }
};

static void PspSetSound(int Channel,int Type);
static void PspDrum(int Type,int Force);
static void PspSetChannels(int Volume,int Switch);
static const signed char *PspGetWave(int Channel);
static void PspSetWave(int Channel,const signed char *Data,int Length,int Rate);
static void PspSound(int Channel,int Freq,int Volume);
void AudioCallback(void* buf, unsigned int *length, void *userdata);

/** StopSound() **********************************************/
/** Temporarily suspend sound.                              **/
/*************************************************************/
void StopSound(void) { pspAudioSetChannelCallback(0, 0, 0); }

/** ResumeSound() ********************************************/
/** Resume sound after StopSound().                         **/
/*************************************************************/
void ResumeSound(void) { pspAudioSetChannelCallback(0, AudioCallback, 0); }

/** InitSound() **********************************************/
/** Initialize sound. Returns 0 on failure, effective rate  **/
/** otherwise. Special cases of rate argument: 0 - return 1 **/
/** and be silent                                           **/
/*************************************************************/
unsigned int InitSound(unsigned int Rate,unsigned int Delay)
{
  int J;

  /* If initialized, shut down */
  TrashSound();

  /* Set driver functions */
  SndDriver.SetSound    = PspSetSound;
  SndDriver.Drum        = PspDrum;
  SndDriver.SetChannels = PspSetChannels;
  SndDriver.Sound       = PspSound;
  SndDriver.SetWave     = PspSetWave;
  SndDriver.GetWave     = PspGetWave;

  /* Reset channels */
  for(J=0;J<SND_CHANNELS;J++)
  {
    CH[J].Freq   = 0;
    CH[J].Volume = 0;
    CH[J].Count  = 0;
  }

  /* Reset variables */
  NoiseGen = 1;
  SndRate  = 0;

  /* Only 44100 supported */
  if(Rate != 44100) return(1);
  SndRate = Rate;

  /* Set current instrments (if reopening device) */
  for(J=0;J<SND_CHANNELS;J++) PspSetSound(J,CH[J].Type);

  ResumeSound();

  /* Done */
  return(SndRate);
}

/** PspSound() ***********************************************/
/** Set sound volume and frequency for a given channel.     **/
/*************************************************************/
void PspSound(int Channel,int Freq,int Volume)
{
  if(!SndRate||(Channel<0)||(Channel>=SND_CHANNELS)) return;
  Volume=Volume<0? 0:Volume>255? 255:Volume;

  /* Store frequency and volume */
  CH[Channel].Freq   = Freq;
  CH[Channel].Volume = Volume;
}

/** TrashSound() *********************************************/
/** Close all devices and free all allocated resources.     **/
/*************************************************************/
void TrashSound(void)
{
  if(SndRate)
  {
    /* Shutdown wave audio */
    StopSound();

    /* Done */
    SndRate=0;
  }
}

/** PspSetSound() ********************************************/
/** Set sound type for a given channel.                     **/
/*************************************************************/
void PspSetSound(int Channel,int Type)
{
  if(!SndRate||(Channel<0)||(Channel>=SND_CHANNELS)) return;

  CH[Channel].Type=Type;
}

/** PspDrum() ************************************************/
/** Hit a drum of a given type with given force.            **/
/*************************************************************/
void PspDrum(int Type,int Force)
{
  if(!SndRate) return;
  Force=Force<0? 0:Force>255? 255:Force;
}

/** PspSetChannels() *****************************************/
/** Set overall sound volume and switch channels on/off.    **/
/*************************************************************/
void PspSetChannels(int Volume,int Switch)
{
  if(!SndRate) return;
  Volume=Volume<0? 0:Volume>255? 255:Volume;
  Switch&=(1<<SND_CHANNELS)-1;

  MasterVolume=Volume;
  MasterSwitch=Switch;
}

/** GetWave() ************************************************/
/** Get current read position for the buffer set with the   **/
/** SetWave() call. Returns 0 if no buffer has been set, or **/
/** if there is no playrate set (i.e. wave is instrument).  **/
/*************************************************************/
const signed char *PspGetWave(int Channel)
{
  /* Channel has to be valid */
  if((Channel<0)||(Channel>=SND_CHANNELS)) return(0);
  /* Return current read position */
  return(
    CH[Channel].Rate&&(CH[Channel].Type==SND_WAVE)?
    CH[Channel].Data+CH[Channel].Pos:0
  );
}


/** SetWave() ************************************************/
/** Set waveform for a given channel. The channel will be   **/
/** marked with sound type SND_WAVE. Set Rate=0 if you want **/
/** waveform to be an instrument or set it to the waveform  **/
/** own playback rate.                                      **/
/*************************************************************/
void PspSetWave(int Channel,const signed char *Data,int Length,int Rate)
{
  /* Channel has to be valid */
  if((Channel<0)||(Channel>=SND_CHANNELS)) return;

  /* Set channel parameters */
  CH[Channel].Type   = SND_WAVE;
  CH[Channel].Length = Length;
  CH[Channel].Rate   = Rate;
  CH[Channel].Data   = Data;
  CH[Channel].Pos    = 0;
  CH[Channel].Count  = 0;
}

/** AudioCallback() ******************************************/
/** Called by the system to render sound                    **/
/*************************************************************/
void AudioCallback(void* buf, unsigned int *length, void *userdata)
{
  unsigned int Samples = *length;
  register int N,J,K,I,L,L1,L2,V,A1,A2;

  /* Reset some variables to keep compiler happy */
  L=N=A2=0;

  /* Truncate to available space */
  J       = *length;
  Samples = Samples>SND_BUFSIZE? SND_BUFSIZE:Samples;
  Samples = Samples>J? J:Samples;
  if(!Samples) return;

  /* Clear mixing buffer */
  for(J=0;J<Samples;J++) Wave[J]=0;

  /* Waveform generator */
  for(J=0;J<SND_CHANNELS;J++)
    if(CH[J].Freq&&(V=CH[J].Volume)&&(MasterSwitch&(1<<J)))
      switch(CH[J].Type)
      {
        case SND_WAVE: /* Custom Waveform */
          /* Waveform data must have correct length! */
          if(CH[J].Length<=0) break;
          /* Start counting */
          K  = CH[J].Rate>0? (SndRate<<15)/CH[J].Freq/CH[J].Rate
                           : (SndRate<<15)/CH[J].Freq/CH[J].Length;
          L1 = CH[J].Pos%CH[J].Length;
          L2 = CH[J].Count;
          A1 = CH[J].Data[L1]*V;
          /* If expecting interpolation... */
          if(L2<K)
          {
            /* Compute interpolation parameters */
            A2 = CH[J].Data[(L1+1)%CH[J].Length]*V;
            L  = (L2>>15)+1;
            N  = ((K-(L2&0x7FFF))>>15)+1;
          }
          /* Add waveform to the buffer */
          for(I=0;I<Samples;I++)
            if(L2<K)
            {
              /* Interpolate linearly */
              Wave[I]+=A1+L*(A2-A1)/N;
              /* Next waveform step */
              L2+=0x8000;
              /* Next interpolation step */
              L++;
            }
            else
            {
              L1 = (L1+L2/K)%CH[J].Length;
              L2 = (L2%K)+0x8000;
              A1 = CH[J].Data[L1]*V;
              Wave[I]+=A1;
              /* If expecting interpolation... */
              if(L2<K)
              {
                /* Compute interpolation parameters */
                A2 = CH[J].Data[(L1+1)%CH[J].Length]*V;
                L  = 1;
                N  = ((K-L2)>>15)+1;
              }
            }
          /* End counting */
          CH[J].Pos   = L1;
          CH[J].Count = L2;
          break;

        case SND_NOISE: /* White Noise */
          /* For high frequencies, recompute volume */
          if(CH[J].Freq<=SndRate) K=0x10000*CH[J].Freq/SndRate;
          else { V=V*SndRate/CH[J].Freq;K=0x10000; }
          L1=CH[J].Count;
          for(I=0;I<Samples;I++)
          {
            L1+=K;
            if(L1&0xFFFF0000)
            {
              if((NoiseGen<<=1)&0x80000000) NoiseGen^=0x08000001;
              L1&=0xFFFF;
            }
            Wave[I]+=(NoiseGen&1? 127:-128)*V;
          }
          CH[J].Count=L1;
          break;

        case SND_MELODIC:  /* Melodic Sound   */
        case SND_TRIANGLE: /* Triangular Wave */
        default:           /* Default Sound   */
          /* Do not allow frequencies that are too high */
          if(CH[J].Freq>=SndRate/2) break;
          K=0x10000*CH[J].Freq/SndRate;
          L1=CH[J].Count;
          for(I=0;I<Samples;I++,L1+=K)
            Wave[I]+=(L1&0x8000? 127:-128)*V;
          CH[J].Count=L1&0xFFFF;
          break;
      }

  sample_t *OutBuf = (sample_t*)buf;

  /* Mix and convert waveforms */
  for(J=0;J<Samples;J++)
  {
    I           = (Wave[J]*MasterVolume)/255;
    I           = I>32767? 32767:I<-32768? -32768:I;
    OutBuf[J].l = OutBuf[J].r = I;
  }

  *length = Samples;
}
