/** PSP helper library ***************************************/
/**                                                         **/
/**                          kybd.c                         **/
/**                                                         **/
/** This file contains virtual keyboard implementation      **/
/** routines                                                **/
/**                                                         **/
/** Copyright (C) Akop Karapetyan 2007                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#include <time.h>
#include <psptypes.h>
#include <psprtc.h>

#include "video.h"
#include "image.h"
#include "font.h"
#include "kybd.h"

struct KeyboardButton
{
  char *Caption;
  unsigned char Code;
  int X;
  int Y;
  int W;
  int H;
};

struct StickyButton
{
  unsigned char Code;
  int Status;
};

#define PSP_KYBD_BUTTONS   6
#define PSP_KYBD_DELAY     400
#define PSP_KYBD_THRESHOLD 50

#define PSP_KYBD_COLOR_BG       RGBA(0, 0, 0, 0x99)
#define PSP_KYBD_COLOR_FONT     RGBA(0xFF, 0xFF, 0xFF, 0xFF)
#define PSP_KYBD_COLOR_STUCK    RGBA(0xFF, 0, 0, 0x99)
#define PSP_KYBD_COLOR_SELECTED RGBA(0, 0, 0xFF, 0xAA)

static int Selected;
static unsigned char Tapped;
static unsigned char HeldDown;

static u64 PushTime[PSP_KYBD_BUTTONS];
static const int ButtonMap[PSP_KYBD_BUTTONS] = 
{
  PSP_CTRL_UP,
  PSP_CTRL_DOWN,
  PSP_CTRL_LEFT,
  PSP_CTRL_RIGHT,
  PSP_CTRL_CIRCLE,
  PSP_CTRL_TRIANGLE
};

static const struct KeyboardButton KeyboardLayout[] = 
{
  { "ESC",    0x1B,  0, 0, 40, 30 }, 
  { "F1",     0x12, 45, 0, 40, 30 },  { "F2",   0x13, 85, 0, 40, 30 }, 
  { "F3",     0x14,125, 0, 40, 30 },  { "F4",   0x15,165, 0, 40, 30 }, 
  { "F5",     0x16,205, 0, 40, 30 },  { "STOP", 0x11,250, 0, 40, 30 },
  { "SELECT", 0x0B, 290, 0, 40, 30 }, { "HOME", 0x0C,330, 0, 40, 30 },
  { "INS",    0x0F, 370, 0, 40, 30 }, { "DEL",  0x0E,410, 0, 40, 30 },

  { "~\n`",  0x60,  0, 35, 30, 30 }, 
  { "!\n1",  0x31, 30, 35, 30, 30 }, { "@\n2", 0x32, 60, 35, 30, 30 }, 
  { "#\n3",  0x33, 90, 35, 30, 30 }, { "$\n4", 0x34,120, 35, 30, 30 }, 
  { "%\n5",  0x35,150, 35, 30, 30 }, { "^\n6", 0x36,180, 35, 30, 30 }, 
  { "&\n7",  0x37,210, 35, 30, 30 }, { "*\n8", 0x38,240, 35, 30, 30 }, 
  { "(\n9",  0x39,270, 35, 30, 30 }, { ")\n0", 0x30,300, 35, 30, 30 },
  { "_\n-",  0x2D,330, 35, 30, 30 }, { "+\n=", 0x3D,360, 35, 30, 30 },
  { "BS",    0x08,390, 35, 60, 30 }, 

  { "TAB", 0x09,  0, 65, 40, 30 }, { "Q",     0x51, 40, 65, 30, 30 }, 
  { "W",   0x57, 70, 65, 30, 30 }, { "E",     0x45,100, 65, 30, 30 }, 
  { "R",   0x52,130, 65, 30, 30 }, { "T",     0x54,160, 65, 30, 30 }, 
  { "Y",   0x59,190, 65, 30, 30 }, { "U",     0x55,220, 65, 30, 30 }, 
  { "I",   0x49,250, 65, 30, 30 }, { "O",     0x4F,280, 65, 30, 30 }, 
  { "P",   0x50,310, 65, 30, 30 }, { "{\n[",  0x5B,340, 65, 30, 30 }, 
  { "}\n]",0x5D,370, 65, 30, 30 }, { "|\n\\", 0x5C,400, 65, 50, 30 }, 

  { "CTRL",  0x06,  0, 95, 50, 30 }, { "A",     0x41, 50, 95, 30, 30 }, 
  { "S",     0x53, 80, 95, 30, 30 }, { "D",     0x44,110, 95, 30, 30 }, 
  { "F",     0x46,140, 95, 30, 30 }, { "G",     0x47,170, 95, 30, 30 }, 
  { "H",     0x48,200, 95, 30, 30 }, { "J",     0x4A,230, 95, 30, 30 }, 
  { "K",     0x4B,260, 95, 30, 30 }, { "L",     0x4C,290, 95, 30, 30 }, 
  { ":\n;",  0x3B,320, 95, 30, 30 }, { "\"\n'", 0x27,350, 95, 30, 30 },
  { "RETURN",0x0D,380, 95, 70, 30 }, 

  { "SHIFT", 0x05,  0,125, 60, 30 }, { "Z",    0x5A, 60,125, 30, 30 }, 
  { "X",     0x58, 90,125, 30, 30 }, { "V",    0x56,120,125, 30, 30 }, 
  { "C",     0x43,150,125, 30, 30 }, { "B",    0x42,180,125, 30, 30 }, 
  { "N",     0x4E,210,125, 30, 30 }, { "M",    0x4D,240,125, 30, 30 },
  { "<\n,",  0x2C,270,125, 30, 30 }, { ">\n.", 0x2E,300,125, 30, 30 }, 
  { "?\n/",  0x2F,330,125, 30, 30 }, { "SHIFT",0x05,360,125, 90, 30 }, 

  { "CAPS",  0x0A, 80,155, 30, 30 }, { "GRPH", 0x07,110,155, 30, 30 },
  { " ",     0x20,140,155,180, 30 }, { "CTRY", 0x10,320,155, 30, 30 },

  { NULL, 0, 0, 0, 0 }
};

static struct StickyButton StickyKeys[] = 
{
  { 0x05, 0 }, { 0x06, 0 }, { 0x07, 0 }, { 0x00, 0 }
};
/*
static struct StickyButton ToggleKeys[] = 
{
  { 0x10, 0 }, { 0x0A, 0 }, { 0x00, 0 }
};
*/
static int MatrixWidth;
static int MatrixHeight;

void _pspKybdFilterRepeats(SceCtrlData *pad);

void pspKybdCreate()
{
  int i;
  for (i = 0; i < PSP_KYBD_BUTTONS; i++)
    PushTime[i] = 0;

  Tapped = 0;
  Selected = 0;
  HeldDown = 0;

  /* Determine keyboard matrix bounds */ 
  MatrixWidth = MatrixHeight = 0;
  for (i = 0; KeyboardLayout[i].Caption; i++)
  {
    if (KeyboardLayout[i].X + KeyboardLayout[i].W > MatrixWidth)
      MatrixWidth = KeyboardLayout[i].X + KeyboardLayout[i].W;
    if (KeyboardLayout[i].Y + KeyboardLayout[i].H > MatrixHeight)
      MatrixHeight = KeyboardLayout[i].Y + KeyboardLayout[i].H;
  }
}

void pspKybdDestroy()
{
}

void pspKybdReinit(int(*Callback)(unsigned int code))
{
  int i;
  for (i = 0; i < PSP_KYBD_BUTTONS; i++)
    PushTime[i] = 0;

  Tapped = 0;
  HeldDown = 0;

  for (i = 0; StickyKeys[i].Code; i++)
    StickyKeys[i].Status = Callback(StickyKeys[i].Code);
}

void pspKybdShow()
{
  /* TODO: Keyboard rendering needs optimization */
  /* It's too slow */
  const struct KeyboardButton *button;
  int off_x, off_y, txt_w, txt_h, i, j, fh;

  fh = pspFontGetLineHeight(&PspStockFont);
  off_x = SCR_WIDTH / 2 - MatrixWidth / 2;
  off_y = SCR_HEIGHT / 2 - MatrixHeight / 2;

  for (i = 0; KeyboardLayout[i].Caption; i++)
  {
    for (j = 0; StickyKeys[j].Code; j++)
      if (KeyboardLayout[i].Code == StickyKeys[j].Code)
        break;

    button = &KeyboardLayout[i];
    pspVideoFillRect(off_x + button->X, 
      off_y + button->Y, 
      off_x + button->X + button->W, 
      off_y + button->Y + button->H,
      (i == Selected) 
        ? PSP_KYBD_COLOR_SELECTED 
        : (StickyKeys[j].Code && StickyKeys[j].Status) 
          ? PSP_KYBD_COLOR_STUCK 
          : PSP_KYBD_COLOR_BG);

    pspVideoDrawRect(off_x + button->X, 
      off_y + button->Y, 
      off_x + button->X + button->W, 
      off_y + button->Y + button->H,
      PSP_VIDEO_GRAY);

    txt_w = pspFontGetTextWidth(&PspStockFont, button->Caption);
    txt_h = pspFontGetTextHeight(&PspStockFont, button->Caption);

    pspVideoPrint(&PspStockFont, (off_x + button->X + button->W / 2) - txt_w / 2, 
      (off_y + button->Y + button->H / 2) - txt_h / 2,
      button->Caption, PSP_KYBD_COLOR_FONT);
  }

  char buttons[] = " \026\242\020 Tap button\t\026\243\020 Unstick all sticky buttons ";
  txt_w = pspFontGetTextWidth(&PspStockFont, buttons);
  off_x = SCR_WIDTH / 2 - txt_w / 2;
  off_y = SCR_HEIGHT - fh;

  pspVideoFillRect(off_x, off_y, off_x + txt_w, off_y + fh, PSP_KYBD_COLOR_BG);
  pspVideoPrint(&PspStockFont, off_x, off_y, buttons, PSP_VIDEO_WHITE);
}

void _pspKybdFilterRepeats(SceCtrlData *pad)
{
  SceCtrlData p = *pad;
  int i;
  u64 tick;
  u32 tick_res;

  /* Get current tick count */
  sceRtcGetCurrentTick(&tick);
  tick_res = sceRtcGetTickResolution();

  /* Check each button */
  for (i = 0; i < PSP_KYBD_BUTTONS; i++)
  {
    if (p.Buttons & ButtonMap[i])
    {
      if (!PushTime[i] || tick >= PushTime[i])
      {
        /* Button was pushed for the first time, or time to repeat */
        pad->Buttons |= ButtonMap[i];
        /* Compute next press time */
        PushTime[i] = tick + ((PushTime[i]) ? PSP_KYBD_THRESHOLD : PSP_KYBD_DELAY)
          * (tick_res / 1000);
      }
      else
      {
        /* No need to repeat yet */
        pad->Buttons &= ~ButtonMap[i];
      }
    }
    else
    {
      /* Button was released */
      pad->Buttons &= ~ButtonMap[i];
      PushTime[i] = 0;        
    }
  }
}

int pspKybdNavigate(SceCtrlData *pad, void(*Callback)(unsigned char, int))
{
  int i, dirty = 0;

  _pspKybdFilterRepeats(pad);

  if (Tapped)
  {
    Callback(Tapped, 0);
    Tapped = 0;
  }

  if (pad->Buttons & PSP_CTRL_RIGHT)
  {
    if (KeyboardLayout[Selected + 1].Caption)
    {
      Selected++;
      dirty = 1;
    }
  }
  else if (pad->Buttons & PSP_CTRL_LEFT)
  {
    if (Selected > 0) 
    {
      Selected--; 
      dirty = 1;
    }
  }
  else if (pad->Buttons & PSP_CTRL_DOWN)
  {
    /* Find first button on the next row */
    for (i = Selected + 1; KeyboardLayout[i].Caption && KeyboardLayout[i].Y == KeyboardLayout[Selected].Y; i++);

    /* Find button below the current one */
    int r = i;
    for (; KeyboardLayout[i].Caption && KeyboardLayout[i].Y == KeyboardLayout[r].Y; i++)
      if (KeyboardLayout[i].X + KeyboardLayout[i].W / 2 >= KeyboardLayout[Selected].X + KeyboardLayout[Selected].W / 2) 
        break;

    Selected = (KeyboardLayout[i].Caption && KeyboardLayout[r].Y == KeyboardLayout[i].Y) ? i : i - 1;
    dirty = 1;
  }
  else if (pad->Buttons & PSP_CTRL_UP)
  {
    /* Find first button on the previous row */
    for (i = Selected - 1; i >= 0 && KeyboardLayout[i].Y == KeyboardLayout[Selected].Y; i--);

    /* Find button above the current one */
    int r = i;
    for (; i >= 0 && KeyboardLayout[i].Y == KeyboardLayout[r].Y; i--)
      if (KeyboardLayout[i].X + KeyboardLayout[i].W / 2 <= KeyboardLayout[Selected].X + KeyboardLayout[Selected].W / 2) 
        break;

    Selected = (i >= 0 && KeyboardLayout[r].Y == KeyboardLayout[i].Y) ? i : i + 1;
    dirty = 1;
  }

  if (pad->Buttons & PSP_CTRL_CIRCLE)
  {
    for (i = 0; StickyKeys[i].Code; i++)
    {
      /* Sticky key; toggle status */
      if (KeyboardLayout[Selected].Code == StickyKeys[i].Code)
      {
        StickyKeys[i].Status = !StickyKeys[i].Status;
        Callback(StickyKeys[i].Code, StickyKeys[i].Status);
        break;
      }
    }

    /* Not a sticky key */
    if (!StickyKeys[i].Code)
    {
      Tapped = KeyboardLayout[Selected].Code;
      Callback(Tapped, 1);
    }

    dirty = 1;
  }
  else if (pad->Buttons & PSP_CTRL_TRIANGLE)
  {
    /* Unset all sticky keys */
    for (i = 0; StickyKeys[i].Code; i++)
    {
      StickyKeys[i].Status = 0;
      Callback(StickyKeys[i].Code, 0);
    }

    dirty = 1;
  }

  /* TODO: Unset used buttons */
  for (i = 0; i < PSP_KYBD_BUTTONS; i++)
    pad->Buttons &= ~ButtonMap[i];

  return dirty;
}
