/** PSP helper library ***************************************/
/**                                                         **/
/**                          png.c                          **/
/**                                                         **/
/** This file contains PNG loading/saving routines. It is   **/
/** based on, and adapted from NesterJ v1.11 code by Ruka   **/
/**                                                         **/
/** Copyright (C) Akop Karapetyan 2007                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#include <png.h>
#include <stdio.h>
#include <stdlib.h>
#include <pspkernel.h>

#include "png.h"
#include "video.h"

typedef unsigned char byte;

PspImage* pspPngLoadImage(const char *path)
{
  FILE *fp = fopen(path,"rb");
  if(!fp) return NULL;

  PspImage *image = pspPngLoadImageOpen(fp);
  fclose(fp);

  return image;
}

int pspPngSaveImage(const char *path, PspImage* image)
{
  FILE *fp = fopen( path, "wb" );
	if (!fp) return 0;

  int stat = pspPngSaveImageOpen(fp, image);
  fclose(fp);

  return stat;
}

PspImage* pspPngLoadImageOpen(FILE *fp)
{
  const size_t nSigSize = 8;
  byte signature[nSigSize];
  if (fread(signature, sizeof(byte), nSigSize, fp) != nSigSize)
    return 0;

  if (!png_check_sig(signature, nSigSize))
    return 0;

  png_struct *pPngStruct = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if(!pPngStruct)
    return 0;

  png_info *pPngInfo = png_create_info_struct(pPngStruct);
  if(!pPngInfo)
  {
    png_destroy_read_struct(&pPngStruct, NULL, NULL);
    return 0;
  }

  if (setjmp(pPngStruct->jmpbuf))
  {
    png_destroy_read_struct(&pPngStruct, NULL, NULL);
    return 0;
  }

  png_init_io(pPngStruct, fp);
  png_set_sig_bytes(pPngStruct, nSigSize);
  png_read_png(pPngStruct, pPngInfo,
    PNG_TRANSFORM_STRIP_16 | PNG_TRANSFORM_PACKING |
    PNG_TRANSFORM_EXPAND | PNG_TRANSFORM_BGR , NULL);

  png_uint_32 width = pPngInfo->width;
  png_uint_32 height = pPngInfo->height;
  int color_type = pPngInfo->color_type;

  PspImage *image;

  if (!(image = pspImageCreate(width, height)))
  {
    png_destroy_read_struct(&pPngStruct, &pPngInfo, NULL);
    return 0;
  }

  png_byte **pRowTable = pPngInfo->row_pointers;
  unsigned int x, y;
  byte r, g, b;
  unsigned short *out = image->Pixels;

  for (y=0; y<height; y++)
  {
    png_byte *pRow = pRowTable[y];
    for (x=0; x<width; x++)
    {
      switch(color_type)
      {
        case PNG_COLOR_TYPE_GRAY:
          r = g = b = *pRow++;
          break;
        case PNG_COLOR_TYPE_GRAY_ALPHA:
          r = g = b = pRow[0];
          pRow += 2;
          break;
        case PNG_COLOR_TYPE_RGB:
          b = pRow[0];
          g = pRow[1];
          r = pRow[2];
          pRow += 3;
          break;
        case PNG_COLOR_TYPE_RGB_ALPHA:
          b = pRow[0];
          g = pRow[1];
          r = pRow[2];
          pRow += 4;
          break;
        default:
          r = g = b = 0;
          break;
      }

      *out++=RGB16(r,g,b);
    }
  }

  png_destroy_read_struct(&pPngStruct, &pPngInfo, NULL);

  return image;
}

int pspPngSaveImageOpen(FILE *fp, PspImage* image)
{
  unsigned char bitmap[image->Height][image->Width][3];
  int i, j, width, height;
  unsigned char r, g, b;
  unsigned short *pixel = image->Pixels;

  width = image->Width;
  height = image->Height;

  for (i = 0; i < height; i++)
  {
    for (j = 0; j < width; j++, pixel++)
    {
      r = (*pixel) & 0x0f;
      g = (*pixel >> 4) & 0x0f;
      b = (*pixel >> 8) & 0x0f;
      bitmap[i][j][0] = r * 0xff / 0x0f;
      bitmap[i][j][1] = g * 0xff / 0x0f;
      bitmap[i][j][2] = b * 0xff / 0x0f;
    }
  }

  unsigned char *in = (unsigned char *)bitmap;
  
  png_struct *pPngStruct = png_create_write_struct( PNG_LIBPNG_VER_STRING,
    NULL, NULL, NULL );

	if (!pPngStruct)
		return 0;

  png_info *pPngInfo = png_create_info_struct( pPngStruct );
	if (!pPngInfo)
	{
		png_destroy_write_struct( &pPngStruct, NULL );
		return 0;
	}

	png_byte **buf = (png_byte**)malloc(height*sizeof(png_byte*));
	if (!buf)
	{
    png_destroy_write_struct( &pPngStruct, &pPngInfo );
		return 0;
	}

	unsigned int y;
	for(y=0; y<height; y++)
		buf[y] = (byte*)&in[y*width*3];

  if (setjmp( pPngStruct->jmpbuf ))
  {
    free(buf);
    png_destroy_write_struct( &pPngStruct, &pPngInfo );
    return 0;
  }

  png_init_io( pPngStruct, fp );
  png_set_IHDR( pPngStruct, pPngInfo, width, height, 8, 
    PNG_COLOR_TYPE_RGB,
    PNG_INTERLACE_NONE,
    PNG_COMPRESSION_TYPE_DEFAULT,
    PNG_FILTER_TYPE_DEFAULT);
  png_write_info( pPngStruct, pPngInfo );
  png_write_image( pPngStruct, buf );
  png_write_end( pPngStruct, pPngInfo );

  png_destroy_write_struct( &pPngStruct, &pPngInfo );
	free(buf);

  return 1;
}
